@@ -9,71 +9,70 @@ from . import errno
 from .utils import ascii_bytes, maybe_str, parse_name
 from ._version import __version__
 
-import collections.abc
 import socket
 import math
 import functools
 import sys
+from collections.abc import Callable, Iterable
+from typing import Any, Optional, Union
 
+IP4 = tuple[str, int]
+IP6 = tuple[str, int, int, int]
 
-exported_pycares_symbols = [
-    # Flag values
-    'ARES_FLAG_USEVC',
-    'ARES_FLAG_PRIMARY',
-    'ARES_FLAG_IGNTC',
-    'ARES_FLAG_NORECURSE',
-    'ARES_FLAG_STAYOPEN',
-    'ARES_FLAG_NOSEARCH',
-    'ARES_FLAG_NOALIASES',
-    'ARES_FLAG_NOCHECKRESP',
+# Flag values
+ARES_FLAG_USEVC = _lib.ARES_FLAG_USEVC
+ARES_FLAG_PRIMARY = _lib.ARES_FLAG_PRIMARY
+ARES_FLAG_IGNTC = _lib.ARES_FLAG_IGNTC
+ARES_FLAG_NORECURSE = _lib.ARES_FLAG_NORECURSE
+ARES_FLAG_STAYOPEN = _lib.ARES_FLAG_STAYOPEN
+ARES_FLAG_NOSEARCH = _lib.ARES_FLAG_NOSEARCH
+ARES_FLAG_NOALIASES = _lib.ARES_FLAG_NOALIASES
+ARES_FLAG_NOCHECKRESP = _lib.ARES_FLAG_NOCHECKRESP
+ARES_FLAG_EDNS = _lib.ARES_FLAG_EDNS
+ARES_FLAG_NO_DFLT_SVR = _lib.ARES_FLAG_NO_DFLT_SVR
 
-    # Nameinfo flag values
-    'ARES_NI_NOFQDN',
-    'ARES_NI_NUMERICHOST',
-    'ARES_NI_NAMEREQD',
-    'ARES_NI_NUMERICSERV',
-    'ARES_NI_DGRAM',
-    'ARES_NI_TCP',
-    'ARES_NI_UDP',
-    'ARES_NI_SCTP',
-    'ARES_NI_DCCP',
-    'ARES_NI_NUMERICSCOPE',
-    'ARES_NI_LOOKUPHOST',
-    'ARES_NI_LOOKUPSERVICE',
-    'ARES_NI_IDN',
-    'ARES_NI_IDN_ALLOW_UNASSIGNED',
-    'ARES_NI_IDN_USE_STD3_ASCII_RULES',
+# Nameinfo flag values
+ARES_NI_NOFQDN = _lib.ARES_NI_NOFQDN
+ARES_NI_NUMERICHOST = _lib.ARES_NI_NUMERICHOST
+ARES_NI_NAMEREQD = _lib.ARES_NI_NAMEREQD
+ARES_NI_NUMERICSERV = _lib.ARES_NI_NUMERICSERV
+ARES_NI_DGRAM = _lib.ARES_NI_DGRAM
+ARES_NI_TCP = _lib.ARES_NI_TCP
+ARES_NI_UDP = _lib.ARES_NI_UDP
+ARES_NI_SCTP = _lib.ARES_NI_SCTP
+ARES_NI_DCCP = _lib.ARES_NI_DCCP
+ARES_NI_NUMERICSCOPE = _lib.ARES_NI_NUMERICSCOPE
+ARES_NI_LOOKUPHOST = _lib.ARES_NI_LOOKUPHOST
+ARES_NI_LOOKUPSERVICE = _lib.ARES_NI_LOOKUPSERVICE
+ARES_NI_IDN = _lib.ARES_NI_IDN
+ARES_NI_IDN_ALLOW_UNASSIGNED = _lib.ARES_NI_IDN_ALLOW_UNASSIGNED
+ARES_NI_IDN_USE_STD3_ASCII_RULES = _lib.ARES_NI_IDN_USE_STD3_ASCII_RULES
 
-    # Bad socket
-    'ARES_SOCKET_BAD',
-]
+# Bad socket
+ARES_SOCKET_BAD = _lib.ARES_SOCKET_BAD
 
-for symbol in exported_pycares_symbols:
-    globals()[symbol] = getattr(_lib, symbol)
-
-
-exported_pycares_symbols_map = {
-    # Query types
-    "QUERY_TYPE_A" : "T_A",
-    "QUERY_TYPE_AAAA" : "T_AAAA",
-    "QUERY_TYPE_ANY" : "T_ANY",
-    "QUERY_TYPE_CNAME" : "T_CNAME",
-    "QUERY_TYPE_MX" : "T_MX",
-    "QUERY_TYPE_NAPTR" : "T_NAPTR",
-    "QUERY_TYPE_NS" : "T_NS",
-    "QUERY_TYPE_PTR" : "T_PTR",
-    "QUERY_TYPE_SOA" : "T_SOA",
-    "QUERY_TYPE_SRV" : "T_SRV",
-    "QUERY_TYPE_TXT" : "T_TXT",
-}
-
-for k, v in exported_pycares_symbols_map.items():
-    globals()[k] = getattr(_lib, v)
-
-
-globals()['ARES_VERSION'] = maybe_str(_ffi.string(_lib.ares_version(_ffi.NULL)))
+# Query types
+QUERY_TYPE_A = _lib.T_A
+QUERY_TYPE_AAAA = _lib.T_AAAA
+QUERY_TYPE_ANY = _lib.T_ANY
+QUERY_TYPE_CAA = _lib.T_CAA
+QUERY_TYPE_CNAME = _lib.T_CNAME
+QUERY_TYPE_MX = _lib.T_MX
+QUERY_TYPE_NAPTR = _lib.T_NAPTR
+QUERY_TYPE_NS = _lib.T_NS
+QUERY_TYPE_PTR = _lib.T_PTR
+QUERY_TYPE_SOA = _lib.T_SOA
+QUERY_TYPE_SRV = _lib.T_SRV
+QUERY_TYPE_TXT = _lib.T_TXT
 
+# Query classes
+QUERY_CLASS_IN = _lib.C_IN
+QUERY_CLASS_CHAOS = _lib.C_CHAOS
+QUERY_CLASS_HS = _lib.C_HS
+QUERY_CLASS_NONE = _lib.C_NONE
+QUERY_CLASS_ANY = _lib.C_ANY
 
+ARES_VERSION = maybe_str(_ffi.string(_lib.ares_version(_ffi.NULL)))
 PYCARES_ADDRTTL_SIZE = 256
 
 
@@ -124,13 +123,13 @@ def _query_cb(arg, status, timeouts, abuf, alen):
     if status == _lib.ARES_SUCCESS:
         if query_type == _lib.T_ANY:
             result = []
-            for qtype in (_lib.T_A, _lib.T_AAAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT):
+            for qtype in (_lib.T_A, _lib.T_AAAA, _lib.T_CAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT):
                 r, status = parse_result(qtype, abuf, alen)
                 if status not in (None, _lib.ARES_ENODATA, _lib.ARES_EBADRESP):
                     result = None
                     break
                 if r is not None:
-                    if isinstance(r, collections.abc.Iterable):
+                    if isinstance(r, Iterable):
                         result.extend(r)
                     else:
                         result.append(r)
@@ -143,6 +142,19 @@ def _query_cb(arg, status, timeouts, abuf, alen):
 
     callback(result, status)
 
+@_ffi.def_extern()
+def _addrinfo_cb(arg, status, timeouts, res):
+    callback = _ffi.from_handle(arg)
+    _global_set.discard(arg)
+
+    if status != _lib.ARES_SUCCESS:
+        result = None
+    else:
+        result = ares_addrinfo_result(res)
+        status = None
+
+    callback(result, status)
+
 def parse_result(query_type, abuf, alen):
     if query_type == _lib.T_A:
         addrttls = _ffi.new("struct ares_addrttl[]", PYCARES_ADDRTTL_SIZE)
@@ -164,6 +176,20 @@ def parse_result(query_type, abuf, alen):
         else:
             result = [ares_query_aaaa_result(addrttls[i]) for i in range(naddrttls[0])]
             status = None
+    elif query_type == _lib.T_CAA:
+        caa_reply = _ffi.new("struct ares_caa_reply **")
+        parse_status = _lib.ares_parse_caa_reply(abuf, alen, caa_reply)
+        if parse_status != _lib.ARES_SUCCESS:
+            result = None
+            status = parse_status
+        else:
+            result = []
+            caa_reply_ptr = caa_reply[0]
+            while caa_reply_ptr != _ffi.NULL:
+                result.append(ares_query_caa_result(caa_reply_ptr))
+                caa_reply_ptr = caa_reply_ptr.next
+            _lib.ares_free_data(caa_reply[0])
+            status = None
     elif query_type == _lib.T_CNAME:
         host = _ffi.new("struct hostent **")
         parse_status = _lib.ares_parse_a_reply(abuf, alen, host, _ffi.NULL, _ffi.NULL)
@@ -176,7 +202,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_MX:
         mx_reply = _ffi.new("struct ares_mx_reply **")
-        parse_status = _lib.ares_parse_mx_reply(abuf, alen, mx_reply);
+        parse_status = _lib.ares_parse_mx_reply(abuf, alen, mx_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -190,7 +216,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_NAPTR:
         naptr_reply = _ffi.new("struct ares_naptr_reply **")
-        parse_status = _lib.ares_parse_naptr_reply(abuf, alen, naptr_reply);
+        parse_status = _lib.ares_parse_naptr_reply(abuf, alen, naptr_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -204,7 +230,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_NS:
         hostent = _ffi.new("struct hostent **")
-        parse_status = _lib.ares_parse_ns_reply(abuf, alen, hostent);
+        parse_status = _lib.ares_parse_ns_reply(abuf, alen, hostent)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -219,8 +245,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_PTR:
         hostent = _ffi.new("struct hostent **")
-        hostttl = _ffi.new("int*", PYCARES_ADDRTTL_SIZE)
-        parse_status = _lib.ares_parse_ptr_reply(abuf, alen, _ffi.NULL, 0, socket.AF_UNSPEC, hostent, hostttl);
+        parse_status = _lib.ares_parse_ptr_reply(abuf, alen, _ffi.NULL, 0, socket.AF_UNSPEC, hostent)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -231,12 +256,12 @@ def parse_result(query_type, abuf, alen):
             while host.h_aliases[i] != _ffi.NULL:
                 aliases.append(maybe_str(_ffi.string(host.h_aliases[i])))
                 i += 1
-            result = ares_query_ptr_result(host, hostttl[0], aliases)
+            result = ares_query_ptr_result(host, aliases)
             _lib.ares_free_hostent(host)
             status = None
     elif query_type == _lib.T_SOA:
         soa_reply = _ffi.new("struct ares_soa_reply **")
-        parse_status = _lib.ares_parse_soa_reply(abuf, alen, soa_reply);
+        parse_status = _lib.ares_parse_soa_reply(abuf, alen, soa_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -246,7 +271,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_SRV:
         srv_reply = _ffi.new("struct ares_srv_reply **")
-        parse_status = _lib.ares_parse_srv_reply(abuf, alen, srv_reply);
+        parse_status = _lib.ares_parse_srv_reply(abuf, alen, srv_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -260,7 +285,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_TXT:
         txt_reply = _ffi.new("struct ares_txt_ext **")
-        parse_status = _lib.ares_parse_txt_reply_ext(abuf, alen, txt_reply);
+        parse_status = _lib.ares_parse_txt_reply_ext(abuf, alen, txt_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -290,25 +315,27 @@ def parse_result(query_type, abuf, alen):
 
 
 class Channel:
-    __qtypes__ = (_lib.T_A, _lib.T_AAAA, _lib.T_ANY, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT)
+    __qtypes__ = (_lib.T_A, _lib.T_AAAA, _lib.T_ANY, _lib.T_CAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT)
+    __qclasses__ = (_lib.C_IN, _lib.C_CHAOS, _lib.C_HS, _lib.C_NONE, _lib.C_ANY)
 
     def __init__(self,
-                 flags = None,
-                 timeout = None,
-                 tries = None,
-                 ndots = None,
-                 tcp_port = None,
-                 udp_port = None,
-                 servers = None,
-                 domains = None,
-                 lookups = None,
-                 sock_state_cb = None,
-                 socket_send_buffer_size = None,
-                 socket_receive_buffer_size = None,
-                 rotate = False,
-                 local_ip = None,
-                 local_dev = None,
-                 resolvconf_path = None):
+                 flags: Optional[int] = None,
+                 timeout: Optional[float] = None,
+                 tries: Optional[int] = None,
+                 ndots: Optional[int] = None,
+                 tcp_port: Optional[int] = None,
+                 udp_port: Optional[int] = None,
+                 servers: Optional[Iterable[Union[str, bytes]]] = None,
+                 domains: Optional[Iterable[Union[str, bytes]]] = None,
+                 lookups: Union[str, bytes, None] = None,
+                 sock_state_cb: Optional[Callable[[int, bool, bool], None]] = None,
+                 socket_send_buffer_size: Optional[int] = None,
+                 socket_receive_buffer_size: Optional[int] = None,
+                 rotate: bool = False,
+                 local_ip: Union[str, bytes, None] = None,
+                 local_dev: Optional[str] = None,
+                 resolvconf_path: Union[str, bytes, None] = None,
+                 event_thread: bool = False) -> None:
 
         channel = _ffi.new("ares_channel *")
         options = _ffi.new("struct ares_options *")
@@ -349,6 +376,8 @@ class Channel:
         if sock_state_cb:
             if not callable(sock_state_cb):
                 raise TypeError("sock_state_cb is not callable")
+            if event_thread:
+                raise RuntimeError("sock_state_cb and event_thread cannot be used together")
 
             userdata = _ffi.new_handle(sock_state_cb)
 
@@ -359,6 +388,14 @@ class Channel:
             options.sock_state_cb_data = userdata
             optmask = optmask |  _lib.ARES_OPT_SOCK_STATE_CB
 
+        if event_thread:
+            if not ares_threadsafety():
+                raise RuntimeError("c-ares is not built with thread safety")
+            if sock_state_cb:
+                raise RuntimeError("sock_state_cb and event_thread cannot be used together")
+            optmask = optmask |  _lib.ARES_OPT_EVENT_THREAD
+            options.evsys = _lib.ARES_EVSYS_DEFAULT
+
         if lookups:
             options.lookups = _ffi.new('char[]', ascii_bytes(lookups))
             optmask = optmask |  _lib.ARES_OPT_LOOKUPS
@@ -395,11 +432,16 @@ class Channel:
         if local_dev:
             self.set_local_dev(local_dev)
 
-    def cancel(self):
+    def cancel(self) -> None:
         _lib.ares_cancel(self._channel[0])
 
+    def reinit(self) -> None:
+        r = _lib.ares_reinit(self._channel[0])
+        if r != _lib.ARES_SUCCESS:
+            raise AresError(r, errno.strerror(r))
+
     @property
-    def servers(self):
+    def servers(self) -> list[str]:
         servers = _ffi.new("struct ares_addr_node **")
 
         r = _lib.ares_get_servers(self._channel[0], servers)
@@ -422,7 +464,7 @@ class Channel:
         return server_list
 
     @servers.setter
-    def servers(self, servers):
+    def servers(self, servers: Iterable[Union[str, bytes]]) -> None:
         c = _ffi.new("struct ares_addr_node[%d]" % len(servers))
         for i, server in enumerate(servers):
             if _lib.ares_inet_pton(socket.AF_INET, ascii_bytes(server), _ffi.addressof(c[i].addr.addr4)) == 1:
@@ -452,7 +494,7 @@ class Channel:
 
         return rfds, wfds
 
-    def process_fd(self, read_fd, write_fd):
+    def process_fd(self, read_fd: int, write_fd: int) -> None:
         _lib.ares_process_fd(self._channel[0], _ffi.cast("ares_socket_t", read_fd), _ffi.cast("ares_socket_t", write_fd))
 
     def timeout(self, t = None):
@@ -474,7 +516,7 @@ class Channel:
 
         return (tv.tv_sec + tv.tv_usec / 1000000.0)
 
-    def gethostbyaddr(self, addr, callback):
+    def gethostbyaddr(self, addr: str, callback: Callable[[Any, int], None]) -> None:
         if not callable(callback):
             raise TypeError("a callable is required")
 
@@ -493,7 +535,7 @@ class Channel:
         _global_set.add(userdata)
         _lib.ares_gethostbyaddr(self._channel[0], address, _ffi.sizeof(address[0]), family, _lib._host_cb, userdata)
 
-    def gethostbyname(self, name, family, callback):
+    def gethostbyname(self, name: str, family: socket.AddressFamily, callback: Callable[[Any, int], None]) -> None:
         if not callable(callback):
             raise TypeError("a callable is required")
 
@@ -501,22 +543,58 @@ class Channel:
         _global_set.add(userdata)
         _lib.ares_gethostbyname(self._channel[0], parse_name(name), family, _lib._host_cb, userdata)
 
-    def query(self, name, query_type, callback):
-        self._do_query(_lib.ares_query, name, query_type, callback)
+    def getaddrinfo(
+        self,
+        host: str,
+        port: Optional[int],
+        callback: Callable[[Any, int], None],
+        family: socket.AddressFamily = 0,
+        type: int = 0,
+        proto: int = 0,
+        flags: int = 0
+    ) -> None:
+        if not callable(callback):
+            raise TypeError("a callable is required")
 
-    def search(self, name, query_type, callback):
-        self._do_query(_lib.ares_search, name, query_type, callback)
+        if port is None:
+            service = _ffi.NULL
+        elif isinstance(port, int):
+            service = str(port).encode('ascii')
+        else:
+            service = ascii_bytes(port)
 
-    def _do_query(self, func, name, query_type, callback):
+        userdata = _ffi.new_handle(callback)
+        _global_set.add(userdata)
+
+        hints = _ffi.new('struct ares_addrinfo_hints*')
+        hints.ai_flags = flags
+        hints.ai_family = family
+        hints.ai_socktype = type
+        hints.ai_protocol = proto
+        _lib.ares_getaddrinfo(self._channel[0], parse_name(host), service, hints, _lib._addrinfo_cb, userdata)
+
+    def query(self, name: str, query_type: str, callback: Callable[[Any, int], None], query_class: Optional[str] = None) -> None:
+        self._do_query(_lib.ares_query, name, query_type, callback, query_class=query_class)
+
+    def search(self, name, query_type, callback, query_class=None):
+        self._do_query(_lib.ares_search, name, query_type, callback, query_class=query_class)
+
+    def _do_query(self, func, name, query_type, callback, query_class=None):
         if not callable(callback):
             raise TypeError('a callable is required')
 
         if query_type not in self.__qtypes__:
             raise ValueError('invalid query type specified')
 
+        if query_class is None:
+            query_class = _lib.C_IN
+
+        if query_class not in self.__qclasses__:
+            raise ValueError('invalid query class specified')
+
         userdata = _ffi.new_handle((callback, query_type))
         _global_set.add(userdata)
-        func(self._channel[0], parse_name(name), _lib.C_IN, query_type, _lib._query_cb, userdata)
+        func(self._channel[0], parse_name(name), query_class, query_type, _lib._query_cb, userdata)
 
     def set_local_ip(self, ip):
         addr4 = _ffi.new("struct in_addr*")
@@ -528,28 +606,30 @@ class Channel:
         else:
             raise ValueError("invalid IP address")
 
-    def getnameinfo(self, ip_port, flags, callback):
+    def getnameinfo(self, address: Union[IP4, IP6], flags: int, callback: Callable[[Any, int], None]) -> None:
         if not callable(callback):
             raise TypeError("a callable is required")
 
-        ip, port = ip_port
-
-        if port < 0 or port > 65535:
-            raise ValueError("port must be between 0 and 65535")
-
-        sa4 = _ffi.new("struct sockaddr_in*")
-        sa6 = _ffi.new("struct sockaddr_in6*")
-
-        if _lib.ares_inet_pton(socket.AF_INET, ascii_bytes(ip), _ffi.addressof(sa4.sin_addr)) == 1:
+        if len(address) == 2:
+            (ip, port) = address
+            sa4 = _ffi.new("struct sockaddr_in*")
+            if _lib.ares_inet_pton(socket.AF_INET, ascii_bytes(ip), _ffi.addressof(sa4.sin_addr)) != 1:
+                raise ValueError("Invalid IPv4 address %r" % ip)
             sa4.sin_family = socket.AF_INET
             sa4.sin_port = socket.htons(port)
             sa = sa4
-        elif _lib.ares_inet_pton(socket.AF_INET6, ascii_bytes(ip), _ffi.addressof(sa6.sin6_addr)) == 1:
+        elif len(address) == 4:
+            (ip, port, flowinfo, scope_id) = address
+            sa6 = _ffi.new("struct sockaddr_in6*")
+            if _lib.ares_inet_pton(socket.AF_INET6, ascii_bytes(ip), _ffi.addressof(sa6.sin6_addr)) != 1:
+                raise ValueError("Invalid IPv6 address %r" % ip)
             sa6.sin6_family = socket.AF_INET6
             sa6.sin6_port = socket.htons(port)
+            sa6.sin6_flowinfo = socket.htonl(flowinfo) # I'm unsure about byteorder here.
+            sa6.sin6_scope_id = scope_id # Yes, without htonl.
             sa = sa6
         else:
-            raise ValueError("invalid IP address")
+            raise ValueError("Invalid address argument")
 
         userdata = _ffi.new_handle(callback)
         _global_set.add(userdata)
@@ -592,6 +672,17 @@ class ares_query_aaaa_result(AresResult):
         self.ttl = ares_addrttl.ttl
 
 
+class  ares_query_caa_result(AresResult):
+    __slots__ = ('critical', 'property', 'value', 'ttl')
+    type = 'CAA'
+
+    def __init__(self, caa):
+        self.critical = caa.critical
+        self.property = maybe_str(_ffi.string(caa.property, caa.plength))
+        self.value = maybe_str(_ffi.string(caa.value, caa.length))
+        self.ttl = -1
+
+
 class ares_query_cname_result(AresResult):
     __slots__ = ('cname', 'ttl')
     type = 'CNAME'
@@ -608,7 +699,7 @@ class ares_query_mx_result(AresResult):
     def __init__(self, mx):
         self.host = maybe_str(_ffi.string(mx.host))
         self.priority = mx.priority
-        self.ttl = mx.ttl
+        self.ttl = -1
 
 
 class ares_query_naptr_result(AresResult):
@@ -622,7 +713,7 @@ class ares_query_naptr_result(AresResult):
         self.service = maybe_str(_ffi.string(naptr.service))
         self.regex = maybe_str(_ffi.string(naptr.regexp))
         self.replacement = maybe_str(_ffi.string(naptr.replacement))
-        self.ttl = naptr.ttl
+        self.ttl = -1
 
 
 class ares_query_ns_result(AresResult):
@@ -638,10 +729,10 @@ class ares_query_ptr_result(AresResult):
     __slots__ = ('name', 'ttl', 'aliases')
     type = 'PTR'
 
-    def __init__(self, hostent, ttl, aliases):
+    def __init__(self, hostent, aliases):
         self.name = maybe_str(_ffi.string(hostent.h_name))
-        self.ttl = ttl
         self.aliases = aliases
+        self.ttl = -1
 
 
 class ares_query_soa_result(AresResult):
@@ -656,7 +747,7 @@ class ares_query_soa_result(AresResult):
         self.retry = soa.retry
         self.expires = soa.expire
         self.minttl = soa.minttl
-        self.ttl = soa.ttl
+        self.ttl = -1
 
 
 class  ares_query_srv_result(AresResult):
@@ -668,7 +759,7 @@ class  ares_query_srv_result(AresResult):
         self.port = srv.port
         self.priority = srv.priority
         self.weight = srv.weight
-        self.ttl = srv.ttl
+        self.ttl = -1
 
 
 class ares_query_txt_result(AresResult):
@@ -677,7 +768,7 @@ class ares_query_txt_result(AresResult):
 
     def __init__(self, txt_chunk):
         self.text = maybe_str(txt_chunk.text)
-        self.ttl = txt_chunk.ttl
+        self.ttl = -1
 
 
 class ares_query_txt_result_chunk(AresResult):
@@ -686,7 +777,7 @@ class ares_query_txt_result_chunk(AresResult):
 
     def __init__(self, txt):
         self.text = _ffi.string(txt.txt)
-        self.ttl = txt.ttl
+        self.ttl = -1
 
 
 # Other result types
@@ -720,7 +811,128 @@ class ares_nameinfo_result(AresResult):
         self.service = maybe_str(_ffi.string(service)) if service != _ffi.NULL else None
 
 
-__all__ = exported_pycares_symbols + list(exported_pycares_symbols_map.keys()) + ['AresError', 'Channel', 'errno', '__version__']
+class ares_addrinfo_node_result(AresResult):
+    __slots__ = ('ttl', 'flags', 'family', 'socktype', 'protocol', 'addr')
 
-del exported_pycares_symbols, exported_pycares_symbols_map
+    def __init__(self, ares_node):
+        self.ttl = ares_node.ai_ttl
+        self.flags = ares_node.ai_flags
+        self.socktype = ares_node.ai_socktype
+        self.protocol = ares_node.ai_protocol
 
+        addr = ares_node.ai_addr
+        assert addr.sa_family == ares_node.ai_family
+        ip = _ffi.new("char []", _lib.INET6_ADDRSTRLEN)
+        if addr.sa_family == socket.AF_INET:
+            self.family = socket.AF_INET
+            s = _ffi.cast("struct sockaddr_in*", addr)
+            if _ffi.NULL != _lib.ares_inet_ntop(s.sin_family, _ffi.addressof(s.sin_addr), ip, _lib.INET6_ADDRSTRLEN):
+                # (address, port) 2-tuple for AF_INET
+                self.addr = (_ffi.string(ip, _lib.INET6_ADDRSTRLEN), socket.ntohs(s.sin_port))
+        elif addr.sa_family == socket.AF_INET6:
+            self.family = socket.AF_INET6
+            s = _ffi.cast("struct sockaddr_in6*", addr)
+            if _ffi.NULL != _lib.ares_inet_ntop(s.sin6_family, _ffi.addressof(s.sin6_addr), ip, _lib.INET6_ADDRSTRLEN):
+                # (address, port, flow info, scope id) 4-tuple for AF_INET6
+                self.addr = (_ffi.string(ip, _lib.INET6_ADDRSTRLEN), socket.ntohs(s.sin6_port), s.sin6_flowinfo, s.sin6_scope_id)
+        else:
+            raise ValueError("invalid sockaddr family")
+
+
+class ares_addrinfo_cname_result(AresResult):
+    __slots__ = ('ttl', 'alias', 'name')
+
+    def __init__(self, ares_cname):
+        self.ttl = ares_cname.ttl
+        self.alias = maybe_str(_ffi.string(ares_cname.alias))
+        self.name = maybe_str(_ffi.string(ares_cname.name))
+
+
+class ares_addrinfo_result(AresResult):
+    __slots__ = ('cnames', 'nodes')
+
+    def __init__(self, ares_addrinfo):
+        self.cnames = []
+        self.nodes = []
+        cname_ptr = ares_addrinfo.cnames
+        while cname_ptr != _ffi.NULL:
+            self.cnames.append(ares_addrinfo_cname_result(cname_ptr))
+            cname_ptr = cname_ptr.next
+        node_ptr = ares_addrinfo.nodes
+        while node_ptr != _ffi.NULL:
+            self.nodes.append(ares_addrinfo_node_result(node_ptr))
+            node_ptr = node_ptr.ai_next
+        _lib.ares_freeaddrinfo(ares_addrinfo)
+
+
+def ares_threadsafety() -> bool:
+    """
+    Check if c-ares was compiled with thread safety support.
+
+    :return: True if thread-safe, False otherwise.
+    :rtype: bool
+    """
+    return bool(_lib.ares_threadsafety())
+
+__all__ = (
+    "ARES_FLAG_USEVC",
+    "ARES_FLAG_PRIMARY",
+    "ARES_FLAG_IGNTC",
+    "ARES_FLAG_NORECURSE",
+    "ARES_FLAG_STAYOPEN",
+    "ARES_FLAG_NOSEARCH",
+    "ARES_FLAG_NOALIASES",
+    "ARES_FLAG_NOCHECKRESP",
+    "ARES_FLAG_EDNS",
+    "ARES_FLAG_NO_DFLT_SVR",
+
+    # Nameinfo flag values
+    "ARES_NI_NOFQDN",
+    "ARES_NI_NUMERICHOST",
+    "ARES_NI_NAMEREQD",
+    "ARES_NI_NUMERICSERV",
+    "ARES_NI_DGRAM",
+    "ARES_NI_TCP",
+    "ARES_NI_UDP",
+    "ARES_NI_SCTP",
+    "ARES_NI_DCCP",
+    "ARES_NI_NUMERICSCOPE",
+    "ARES_NI_LOOKUPHOST",
+    "ARES_NI_LOOKUPSERVICE",
+    "ARES_NI_IDN",
+    "ARES_NI_IDN_ALLOW_UNASSIGNED",
+    "ARES_NI_IDN_USE_STD3_ASCII_RULES",
+
+    # Bad socket
+    "ARES_SOCKET_BAD",
+
+
+    # Query types
+    "QUERY_TYPE_A",
+    "QUERY_TYPE_AAAA",
+    "QUERY_TYPE_ANY",
+    "QUERY_TYPE_CAA",
+    "QUERY_TYPE_CNAME",
+    "QUERY_TYPE_MX",
+    "QUERY_TYPE_NAPTR",
+    "QUERY_TYPE_NS",
+    "QUERY_TYPE_PTR",
+    "QUERY_TYPE_SOA",
+    "QUERY_TYPE_SRV",
+    "QUERY_TYPE_TXT",
+
+    # Query classes
+    "QUERY_CLASS_IN",
+    "QUERY_CLASS_CHAOS",
+    "QUERY_CLASS_HS",
+    "QUERY_CLASS_NONE",
+    "QUERY_CLASS_ANY",
+
+
+    "ARES_VERSION",
+    "AresError",
+    "Channel",
+    "ares_threadsafety",
+    "errno",
+    "__version__"
+)
